/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.IPcept;

import java.io.*;
import java.net.*;

/**
 * The SocketControl class is intended to add functionality to the Java
 * Socket class without writing a new socket implementation.  In particular,
 * SocketControl supports the use of the IPcept facility.  However, other
 * generally useful functions are also included (but should really be in
 * a different package).  This class requires native code support.
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version 1.2
 */

public class SocketControl {

   /**
    * Load the native code.
    */
   static {
      System.loadLibrary("IPceptCtl");
   }

   /**
    * Performs a half-close on the input side of the file or socket 
    * associated with this input stream.  This only works on FileInputStream
    * objects, but it's more convenient to pass an InputStream here.
    * @param is The input stream of the file or socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the input stream is not a 
    *                                FileInputStream or if the shutdown()
    *                                call returns an error
    **/
   public static void shutdown(InputStream is) throws IOException {
      if (is instanceof FileInputStream) {
         shutdownInputStream((FileInputStream) is);
      } else {
         throw new IOException("Invalid operation");
      }
   }

   /**
    * Performs a half-close on the output side of the file or socket 
    * associated with this input stream.  This only works on FileOutputStream
    * objects, but it's more convenient to pass an OutputStream here.
    * @param is The output stream of the file or socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the input stream is not a 
    *                                FileOutputStream or if the shutdown()
    *                                call returns an error
    **/
   public static void shutdown(OutputStream os) throws IOException {
      if (os instanceof FileOutputStream) {
         shutdownOutputStream((FileOutputStream) os);
      } else {
         throw new IOException("Invalid operation");
      }
   }

   /**
    * Performs a half-close on the input side of a socket.
    * @param is The socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the shutdown() call returns an error
    **/
   public static void shutdownInput(Socket s) throws IOException {
      shutdownSock(s, 0);
   }

   /**
    * Performs a half-close on the output side of a socket.
    * @param is The socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the shutdown() call returns an error
    **/
   public static void shutdownOutput(Socket s) throws IOException {
      shutdownSock(s, 1);
   }

   /**
    * Performs a half-close on the input side of the file or socket 
    * associated with this input stream. 
    * @param is The input stream of the file or socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the shutdown() call returns an error
    **/
   protected static native void shutdownInputStream(FileInputStream is)
                                                    throws IOException;

   /**
    * Performs a half-close on the output side of the file or socket 
    * associated with this input stream.
    * @param is The output stream of the file or socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the shutdown() call returns an error
    **/
   protected static native void shutdownOutputStream(FileOutputStream os)
                                                    throws IOException;

   /**
    * Performs a half-close on a socket.
    * @param is The socket to shut down.
    * @param mode As in the shutdown() system call, 0 for input, 1 for 
    *             output, 2 for both.
    * @return No return value
    * @exception java.io.IOException if the shutdown() call returns an error
    **/
   protected static native void shutdownSock(Socket s, int mode) 
                                                    throws IOException;

   /**
    * Creates a socket that is not proxied by IPcept.  Essentially, this
    * is a replacement for the Socket constructor, but the IPCEPT_DONT_PROXY
    * sockopt is set before the connect call is made.  Note that for the
    * superuser, the source port may be able to be a non-local address 
    * (useful for a transparent proxy).
    * @param destAddr the destination address to connect to
    * @param destPort the destination port to connection to
    * @param srcAddr the source address to bind to
    * @param srcPort the source port to bind to
    * @return a new Socket object connected directly to the destination
    *         address and port without a local proxy
    * @exception java.io.IOException if an error occurs while creating, 
    *                                initializing, or connecting the new 
    *                                Socket.
    */
   public static native Socket createUnproxiedSocket(InetAddress destAddr, 
                            int destPort, InetAddress srcAddr, int srcPort)
                                         throws IOException;

   /**
    * Creates a socket that is not proxied by IPcept.  Essentially, this
    * is a replacement for the Socket constructor, but the IPCEPT_DONT_PROXY
    * sockopt is set before the connect call is made.  This method
    * binds the local end of the socket to the local address.
    * @param destAddr the destination address to connect to
    * @param destPort the destination port to connection to
    * @return a new Socket object connected directly to the destination
    *         address and port without a local proxy
    * @exception java.io.IOException if an error occurs while creating, 
    *                                initializing, or connecting the new 
    *                                Socket.
    */
   public static Socket createUnproxiedSocket(InetAddress destAddr, 
             int destPort) throws IOException {
      return createUnproxiedSocket(destAddr, destPort, null, 0);
   }

   /**
    * Get the address of the local endpoint of a socket.  The address is
    * determined directly, using system calls, so no caching in the socket
    * object will affect the outcome.
    * @param s The socket in question.
    * @return a new InetAddress object containing the address that this socket
    *         is bound to
    */
   public static native InetAddress getLocalAddress(Socket s);

   /**
    * Get the port number of the local endpoint of a socket.  The port number 
    * is determined directly, using system calls, so no caching in the socket
    * object will affect the outcome.
    * @param s The socket in question.
    * @return the local port number of the socket
    */
   public static native int getLocalPort(Socket s);

   /**
    * Get the address of the remote endpoint of a socket.  The address is
    * determined directly, using system calls, so no caching in the socket
    * object will affect the outcome.
    * @param s The socket in question.
    * @return a new InetAddress object containing the address that this socket
    *         is connected to
    */
   public static native InetAddress getPeerAddress(Socket s) throws IOException;

   /**
    * Get the port number of the remote endpoint of a socket.  The port number 
    * is determined directly, using system calls, so no caching in the socket
    * object will affect the outcome.
    * @param s The socket in question.
    * @return the remote port number of the socket
    */
   public static native int getPeerPort(Socket s) throws IOException;
}
