/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * IPcept module
 *
 * Author:   Mark Yarvis
 * Date:     June 27, 1998
 * Version:  1.0
 *
 * The IPcept module enables transparent interception of local TCP 
 * connections.  This functionality is very similar to the transparent
 * proxy facility in the firewall, but IPcept allows transparent 
 * interception of local sockets.
 *
 * IPcept implements a set of protocol operations that
 *	(1) allows a bound socket to specify that it will proxy for a 
 *          particular port (through a new socket option).
 *      (2) transparently changes the destination address of a socket
 *	    if the original destination has been proxied.
 *	(3) allows an accepting socket to determine the original destination
 *	    of the client.
 *	(4) allows a connecting socket to masquerade as a different source by
 *	    changing its source address.
 *	(5) reports appropriate addresses for getsockname() and getpeername()
 *
 * These protocol operations are stacked on top of the normal TCP operations.
 * A call to each IPcept protocol operation typically calls the same TCP 
 * operation.
 *
 * Support for protocol stacking is required in the kernel in the form of 
 * the sock_replace() call.  The IPcept module replaces the normal IP
 * proto_ops structure with the IPcept proto_ops structure.  So, each new 
 * IP socket receives the IPcept operation structure.  As soon as it can
 * be determined that the socket will not proxy, be proxied, masquerade, 
 * or be masqueraded, the socket's operations are replaced with the IP 
 * proto_ops, removing IPcept from the picture.
 *
 * Once it is determined that a socket will definately require IPcept 
 * support (for proxying or masquerading), a structure will be created, 
 * allowing IPcept to track the type of support requested.  These structures
 * are kept in a single linked list (this could be changed, in the future,
 * for performance) and accessed using the lookup*() functions.
 *
 * This module can not be unloaded while there exists a socket that is still
 * using the IPcept operations.  In particular, any bound socket (even one
 * not proxying) that was created after loading IPcept will continue to 
 * use the IPcept operations, because of the possibility that, during an
 * accept(), the client will be masquerading.
 *
 *
 * June 22   Code base modified by kfe (Kevin Eustice)
 *           Ported from Linux 2.0 to Linux 2.2.  This involved restructuring 
 *           the driver somewhat.  2.2 requires a net_family structure that is
 *           an array of families. Within each family, there is a family create
 *           function that assigns a proto_ops structure to each created 
 *           socket.  We intercept this create call, and reassign our home 
 *           grown IPcept_proto_ops to the socket.  Note this is based on
 *           Mark's original Conductor code, not Vince's updated UDP IPcept.
 *           Next step is to get UDP integrated back into the code base for
 *           Panda purposes.
 * 
 * July 6th  IPcept_proto_ops replaced with IPcept_tcp_ops structure.  Created 
 *           IPcept_udp_ops structure to complement it.
 *    
 * July 7th  Imported the remaining UDP code, and updated the imported code
 *           to be consistent with 2.2 architectural changes.  Module still 
 *           cannot be unloaded. Must ask Vince why.  
 *
 * July 13th Added bugs to code, and discovered pre-existing ones.  Somewhere in
 *           port to 2.2 we managed to lose the IP address of the proxied host that gets 
 *           embedded in the payload.  Also, Vince and I (Kevin) tried to make the module
 *           removable.  We succeeded to a degree, however now clients segfault when they don't 
 *           exit gracefully, and non managed UDP (ie. NFS, etc) will get upset.
 */

#define MODULE
#define __KERNEL__

#define DEBUG 1


#include <linux/module.h>
#include <linux/poll.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/netdevice.h>
#include <net/sock.h>
#include <net/inet_common.h>

#include <asm/system.h>
#include <linux/proc_fs.h>


#include "IPcept.h"
/* The protocol family structure for IP that we replaced. */
static struct net_proto_family *ipfamily;

/*
 * The list of sockets we manage (usually this only includes the ones that
 * have special attributes). 
 */
static struct cept *cept_socks;

/***********************************************************************
 **                   Manage data structures                          **
 ***********************************************************************/

/*
 * Delete the struct cept associated with the given socket.
 * Returns 1 if object deleted, 0 if object not found.
 */


static __inline__ struct proto_ops *getsockop(struct socket *sock) {
  switch(sock->type) {
  case SOCK_STREAM:
    return &inet_stream_ops;

  case SOCK_DGRAM:
    return &inet_dgram_ops;
    
  default:
    return sock->ops;

  }
}
static int del_cept_socket(struct socket *sock) {
   struct cept *cs;

   /* If there are no elements, return */
   if (cept_socks==NULL) {
      return 0;
   }

   /* If it's the first element, remove it and return */
   if (cept_socks->sock == sock) {
      cs = cept_socks;
      cept_socks = cept_socks->next;
      kfree(cs);
      return 1;
   }

   /* Handle anything after the first element */
   for (cs = cept_socks; cs->next!=NULL; cs = cs->next) {
      if (cs->next->sock == sock) {
         struct cept *temp;
         temp = cs->next;
         cs->next = cs->next->next;
         kfree(temp);
         return 1;
      }
   }

   return 0;
}

/* 
 * Add a new struct cept, associated with the given sock, to the cept_socks 
 * list.
 */
static struct cept *add_socket(struct socket *sock) {
   struct cept *cs = (struct cept *) kmalloc(sizeof(struct cept), GFP_KERNEL);

   cs->flags = 0;
   cs->sock = sock;
   cs->sk = (struct sock *) sock->sk;
   cs->num_ports = 0;
   cs->num_close = 0;
   cs->udp_connected = 0;
   cs->last_recv_dest.sin_family = AF_INET;
   cs->last_recv_dest.sin_addr.s_addr = INADDR_ANY; /* local */
   cs->last_recv_dest.sin_port = htons(0);
   cs->next = cept_socks;
   cept_socks = cs;

   return cs;
}



/*
 * Returns a cept struct that is proxying for a given port (if there is 
 * one), and NULL if there isn't.
 */
static struct cept * lookup_proxy(int type, unsigned short redir) {
   struct cept *cs;

   for (cs = cept_socks; cs!=NULL; cs = cs->next) {
      if (IS_PROXY(cs) && (cs->sock->type == type)) {
         int i;
         for (i=0; i<cs->num_ports; i++) {
            if (cs->redir_port[i] == redir) {
               return cs;
            }
         }
      }
   }
   return NULL;
}

/*
 * Returns the cept struct that is pointing to the supplied socket (if 
 * there is one), and NULL if there isn't.
 */
static struct cept * lookup_socket(struct socket *sock) {
   struct cept *cs;

   for (cs = cept_socks; cs!=NULL; cs = cs->next) {
      if (cs->sock == sock) {
         return cs;
      }
   }
   return NULL;
}

/*
 * Returns the cept struct that is pointing to the socket with the
 * specified address and port number (if there is one), and NULL if 
 * there isn't.  Arguments are given in host byte order.
 */
static struct cept * lookup_address(long addr, unsigned short port) {
   struct cept *cs;

   for (cs = cept_socks; cs!=NULL; cs = cs->next) {
      if ((ntohl(cs->sk->saddr) == addr) && (cs->sk->num == port)) {
         return cs;
      }
   }
   return NULL;
}

/*
 * Adds a port num into cept.close_port arry if it doesn't exist.
 * This is used to keep track which port numbers we sent to,
 * so we can send close notification to the appropriate socket.
 * Returns index of port in cept.close_port if successfull,
 * or a negative error code.
 */
int add_close_port(struct socket *sock, unsigned short num)
{
    int i = 0;
    struct cept *cs = lookup_socket(sock);

    if (cs == NULL)
        cs = add_socket(sock);

    if (cs == NULL)
        return -ENOBUFS;

    SET_HAS_ALT_REMOTE(cs);
    
    while (i < cs->num_close) {
        if (num == cs->close_port[i])
            return i;
        i++;
    }

    if (cs->num_close == MAX_PROXY_PORTS) {
        return -ENOBUFS;
    } else {
        cs->close_port[cs->num_close++] = num;
        return cs->num_close - 1;
    }
}



/*
 * Free up memory set asside for managed sockets.  (Although, by the time
 * this is called, there shouldn't be any.)
 */
static void cleanup_cept_socks() {
   struct cept *cs;
   struct cept *temp;

   cs = cept_socks;
   while (cs!=NULL) {
      temp = cs;
      if(temp->sock->type == SOCK_DGRAM) temp->sock->ops = getsockop(temp->sock);
      cs = cs->next;
      kfree(temp);
   }

   cept_socks = NULL;
}

/***********************************************************************
 **                   Utility functions                               **
 ***********************************************************************/

/*
 * Switch the given socket over to the normal IP ops.  We loose all control
 * of the socket here.
 */
static void unmanage(struct socket * sock) {
#if DEBUG>1
   printk("<1>IPcept (unmanage): unmanaging socket %x\n", sock);
#endif
   sock->ops = getsockop(sock);
   if(sock->type == SOCK_DGRAM) {
     if(lookup_socket(sock)->flags & CEPT_UDP_USED) {
       MOD_DEC_USE_COUNT;
     }
   }
   else {
     MOD_DEC_USE_COUNT;
   }
 }
/*
 * Print out an address in dot notation.  The address is assumed to be in
 * host byte order.
 */
static int sprintAddress(char *buf, unsigned long addr) {
   char *a;

   a = (char *) &addr;

   return sprintf(buf, "%d.%d.%d.%d", (0xff & a[3]), (0xff & a[2]), 
                                      (0xff & a[1]), (0xff & a[0]));
}

/***********************************************************************
 **                   Redirected calls                                **
 ***********************************************************************/

/*
 * Intercept calls to create.  If this isn't a TCP socket, forget it.
 */
static int cept_create(struct socket *sock, int protocol) {
  int tempsocknum;
#if DEBUG>1
  printk("<1>IPcept (cept_create): creating socket %x\n", sock);
#endif


  if(sock->type !=SOCK_DGRAM)  MOD_INC_USE_COUNT;
  
  /* do not intercept non-TCP or UDP sockets */
  if (sock->type == SOCK_DGRAM) { //Is it UDP?
    tempsocknum = ipfamily->create(sock,protocol);
#ifndef TCP_ONLY
    sock->ops = &IPcept_udp_ops; //UDP specific operations
#endif
    return tempsocknum;
  }
  else if (sock->type == SOCK_STREAM) { //Is it TCP?
    tempsocknum = ipfamily->create(sock, protocol);
    sock->ops = &IPcept_tcp_ops; //TCP specific operations
      return tempsocknum;
  }
  else {//If not, we don't care about it.
    unmanage(sock); 
    return ipfamily->create(sock,protocol);
  }
}



/*
 * Intercept dup calls.  Since the new socket inherits IPcept operations
 * from the original socket, we must increment the use count.
 */
static int cept_dup(struct socket *newsock, struct socket *oldsock) {
#if DEBUG>1
   printk("<1>IPcept (cept_dup): creating socket %x from %x\n", 
          newsock, oldsock);
#endif

  if(newsock->type = SOCK_STREAM) MOD_INC_USE_COUNT;
   return (getsockop(oldsock))->dup(newsock, oldsock);
}

/*
 * Intercept release calls.  Free up any memory used and decrement use count.
 */
static int cept_release(struct socket *sock, struct socket *peer) {
#if DEBUG>1
   printk("<1>IPcept (cept_release): releasing socket %x\n",sock);
#endif



   /* in case this socket was managed, delete it */
   del_cept_socket(sock);
   MOD_DEC_USE_COUNT;

   return (getsockop(sock))->release(sock, peer);
}

static int cept_bind(struct socket *sock, struct sockaddr *addr, 
                     int sockaddr_len) {
   struct sockaddr_in *sin=(struct sockaddr_in *)addr;

   if (sockaddr_len < sizeof(struct sockaddr_in)) {
      return -EINVAL;
   }

   if ((sin->sin_addr.s_addr == INADDR_ANY) || 
       (inet_addr_type(sin->sin_addr.s_addr) == RTN_LOCAL)) {
      return (getsockop(sock))->bind(sock, addr, sockaddr_len);
   } else {
      struct cept *csp = lookup_socket(sock);

      /* must be superuser to masquerade */
      if (! oldsuser()) {
         return -EPERM;
      }

      if (csp == NULL) {
         csp = add_socket(sock);
      }

#if DEBUG>0
   printk("<1>IPcept (cept_bind): binding socket %x to foreign address\n",sock);
#endif

      SET_HAS_ALT_LOCAL(csp);
      memcpy(&(csp->local_addr), addr, sockaddr_len);
   }
}

/*
 * Intercept connect calls.  If there is a proxy for our dest port,
 * change the destination address to point to the proxy.  If not,
 * and we're not managed for some other reason (perhaps masquerading),
 * we no longer need to manage this socket.
 */
static int cept_connect(struct socket *sock, struct sockaddr *servaddr,
                        int sockaddr_len, int flags) {
   struct cept *csp = lookup_socket(sock);

   if (sockaddr_len < sizeof(struct sockaddr_in)) {
      return -EINVAL;
   }

   if ( (! csp) || (! DONT_PROXY(csp)) ) {
      struct sockaddr_in *sin=(struct sockaddr_in *)servaddr;
      struct cept *proxy = lookup_proxy(sock->type, ntohs(sin->sin_port));

      if ((proxy!=NULL) && (proxy->sk->num != 0)) {

         if (csp == NULL) {
            csp = add_socket(sock);
         }

         /* save the originally requested address */
         SET_HAS_ALT_REMOTE(csp);
         memcpy(&(csp->remote_addr), sin, sizeof(struct sockaddr_in));

         /* if SOCKADDR_ANY was requested, get address from proxy socket */
         if (csp->remote_addr.sin_addr.s_addr == 0) {
            csp->remote_addr.sin_addr.s_addr = proxy->sk->saddr;
         }

         /* replace address with a local address */
         sin->sin_addr.s_addr = INADDR_ANY;   /* local */
         sin->sin_port = htons(proxy->sk->num);

#if DEBUG>0
         printk("<1>IPcept (cept_connect): redirecting socket %x to %d %d\n",
                sock, ntohl(sin->sin_addr.s_addr), ntohs(sin->sin_port));
#endif
      }
   }

   /*
    * Unless there's some other reason we're managing this socket,
    * we can unmanage it.
    */
   if (! (csp && NEEDS_MANAGEMENT(csp)) ) {
      /* Port not intercepted; use normal operations */
      if (csp) {
         del_cept_socket(sock);
      }
      unmanage(sock);
   }

   return (getsockop(sock))->connect(sock, servaddr, sockaddr_len, flags);
}

static int cept_socketpair(struct socket *sock1, struct socket *sock2) {
   return (getsockop(sock1))->socketpair(sock1, sock2);
}

/*
 * Intercept the accept call.  For the new socket, look for the peer socket 
 * and copy any alternate addresses.  If there's nothing special about this 
 * socket by now, there never will be, unmanage it.
 */
static int cept_accept(struct socket *sock, struct socket *newsock, 
                       int flags) {
   int res = (getsockop(sock))->accept(sock, newsock, flags);
   struct sock *sk=(struct sock *)newsock->sk;
   struct cept *csp = NULL;
   struct cept *peer;

   if (res<0) {
      unmanage(newsock);
      return res;
   }

   peer = lookup_address(ntohl(sk->daddr), ntohs(sk->dport));
   if (peer != NULL) {
      csp = lookup_socket(newsock);
      if (csp == NULL) {
         csp = add_socket(newsock);
      }

      /* is the client masquerading? */
      if (HAS_ALT_LOCAL(peer)) {
         SET_HAS_ALT_REMOTE(csp);
         memcpy(&(csp->remote_addr), &(peer->local_addr), 
                                        sizeof(struct sockaddr_in));
#if DEBUG>1
         printk("<1>IPcept (accept): socket %x accepted connection from masquerading peer\n", sock);
#endif
      }

      /* was the client proxied? */
      if (HAS_ALT_REMOTE(peer)) {
         SET_HAS_ALT_LOCAL(csp);
         memcpy(&(csp->local_addr), &(peer->remote_addr), 
                                        sizeof(struct sockaddr_in));
#if DEBUG>1
         printk("<1>IPcept (accept): socket %x accepted proxied connection\n", sock);
#endif
      }
   }

   /* If we're not managing the new sock now, we no longer need to */
   if (! (csp && NEEDS_MANAGEMENT(csp)) ) {
      if (csp) {
         del_cept_socket(newsock);
      }
      unmanage(newsock);
   }

   return res;
}

/* 
 * If an alternate name of the type requested (peer or not) is requested,
 * use it.  Otherwise, fall through to the normal operation.  Note, this
 * function is also called during the return from accept().
 */
static int cept_getname(struct socket *sock, struct sockaddr *addr,
                        int *sockaddr_len, int peer) {
   struct cept *cp = lookup_socket(sock);

   if ((peer) && (cp != NULL) && HAS_ALT_REMOTE(cp)) {
      *sockaddr_len = sizeof(cp->remote_addr);
      memcpy(addr, &(cp->remote_addr), *sockaddr_len);
      return 0;
   } else if ((! peer) && (cp != NULL) && HAS_ALT_LOCAL(cp)) {
      *sockaddr_len = sizeof(cp->local_addr);
      memcpy(addr, &(cp->local_addr), *sockaddr_len);
      return 0;
   } else {
      return (getsockop(sock))->getname(sock, addr, sockaddr_len, peer);
   }
}

static unsigned int cept_poll(struct file *file, struct socket *sock, poll_table *wait) {
   return (getsockop(sock))->poll(file, sock, wait);
}

static int cept_ioctl(struct socket *sock, unsigned int cmd, 
                      unsigned long arg) {
   return (getsockop(sock))->ioctl(sock, cmd, arg);
}

static int cept_listen(struct socket *sock, int len) {
   return (getsockop(sock))->listen(sock, len);
}

static int cept_shutdown(struct socket *sock, int flags) {
   return (getsockop(sock))->shutdown(sock, flags);
}

/* 
 * Handle SOL_IPCEPT level calls or pass the call along to IP.  Most
 * calls require superuser privlidges.  This call frequently results
 * in the creation of a new cept struct to hold the info.
 */
static int cept_setsockopt(struct socket *sock, int level, int optname,
                           char *optval, int optlen) {
   struct sock *sk=(struct sock *)sock->sk;

#if DEBUG>1
   printk("<1>IPcept (cept_setsockopt): called for %x with %d, %d\n", 
          sock, level, optname);
#endif

   if (level == SOL_IPCEPT) {
      int err=0;
      struct cept *csp = lookup_socket(sock);
      
      /* must be superuser to set most options */
      if (PRIVILIGED_OPTION(optname) && (! oldsuser())) {
	printk("<1>IPcept (cept_setsockopt): returning with bad permissions.\n");
         return -EPERM;
      }

      if (csp == NULL) {
         csp = add_socket(sock);
      }

      if (optname == IPCEPT_INTERCEPT) {
         int val;

         if (optlen < sizeof(int)) {
            err = -EINVAL;
         }

         if (!err) {
            err=verify_area(VERIFY_READ, optval, sizeof(int));
         }

         if(!err) {
	   //val = get_user((int *) optval);
	   get_user(val, (int *) optval); //returns error code, 0 for success
            /* Is there already a proxy for this port? */
            if (lookup_proxy(sock->type, ntohs(val)) != NULL) {
               err = -EADDRINUSE;
               /* fall through for error handling */
            } else {
               /* proxy this port */
               SET_IS_PROXY(csp);
               if (csp->num_ports >= MAX_PROXY_PORTS) {
                  return -EPERM;
               }
               csp->redir_port[(csp->num_ports)++] = ntohs(val);
#if DEBUG>0
               printk("<1>%s: Socket %x proxies port %d to port %d\n",
                      "IPcept (setsockopt)", sock, ntohs(val), sk->num);
#endif
               return 0;  /* success! */
            }
         }
         /* on error, fall through to error handling code */

      } else if (optname == IPCEPT_MASQUERADE) {
         if (optlen < sizeof(struct sockaddr_in)) {
            err = -EINVAL;
         }

         if (!err) {

#if DEBUG>0
   printk("<1>IPcept (cept_setsockopt): masquerading enabled on %x\n",sock);
#endif

            SET_HAS_ALT_LOCAL(csp);
            if(copy_from_user(&(csp->local_addr), optval, 
			      sizeof(struct sockaddr_in))) return -EFAULT;
            return 0;
         }
         /* on error, fall through to error handling code */

      } else if (optname == IPCEPT_DONT_PROXY) {

#if DEBUG>0
   printk("<1>IPcept (cept_setsockopt): proxying disabled on %x\n",sock);
#endif

         SET_DONT_PROXY(csp);
         return 0;

      } else if (optname == IPCEPT_DONT_MASQUERADE) {

#if DEBUG>0
   printk("<1>IPcept (cept_setsockopt): masquerading disabled on %x\n",sock);
#endif

         SET_DONT_MASQUERADE(csp);
         return 0;

      } else if (optname == IPCEPT_UNMANAGE) {

         del_cept_socket(sock);
         unmanage(sock);
         return 0;

      } else {
         err = -ENOPROTOOPT;
         /* fall through to error handling code */
      }

      /* errors fall through to here */
      if (!NEEDS_MANAGEMENT(csp)) {
         del_cept_socket(sock);
      }
      return err;
   } else {
      return (getsockop(sock))->setsockopt(sock, level, optname, optval, optlen);
   }
}

static int cept_getsockopt(struct socket *sock, int level, int optname,
                           char *optval, int *optlen) {
   if ((level == SOL_SOCKET) && (optname == SO_BINDTODEVICE)) {
      struct cept *csp = lookup_socket(sock);
      if ((csp!=NULL) && (HAS_ALT_LOCAL(csp))) {
         struct ifreq req;
	 int err=verify_area(VERIFY_WRITE,optval,sizeof(req));
         if (csp->sk->bound_dev_if && dev_get_by_index(csp->sk->bound_dev_if)) {
            strncpy(req.ifr_name, (dev_get_by_index(csp->sk->bound_dev_if))->name, 
                                           sizeof(req.ifr_name));
            (*(struct sockaddr_in *) &req.ifr_addr).sin_family = csp->sk->family;
            (*(struct sockaddr_in *) &req.ifr_addr).sin_addr.s_addr = 
                                               csp->local_addr.sin_addr.s_addr;
         }

	 if(copy_to_user(optval, &req, sizeof(req))) return -EFAULT;

         return 0;
      }
   }
   return (getsockop(sock))->getsockopt(sock, level, optname, optval, optlen);
}

static int cept_fcntl(struct socket *sock, unsigned int cmd, 
                      unsigned long arg) {
   return (getsockop(sock))->fcntl(sock, cmd, arg);
}

static int cept_sendmsg(struct socket *sock, struct msghdr *msg, int size,          struct scm_cookie *scm) {
   return (getsockop(sock))->sendmsg(sock, msg, size, scm);
}

static int cept_recvmsg(struct socket *sock, struct msghdr *m, int total_len, 
                        int flags, struct scm_cookie *scm) {
   return (getsockop(sock))->recvmsg(sock, m, total_len, flags, scm);
}



/************************* UDP calls ****************************************/

/*
 * Intercept dup calls.  Since the new socket inherits IPcept operations
 * from the original socket, we must increment the use count.
 */
static int udp_dup(struct socket *newsock, struct socket *oldsock) {
  int sock;
  struct cept *csp;
#if DEBUG>1
   printk("<1>IPcept (udp_dup): creating socket %x from %x\n", 
          newsock, oldsock);
#endif

   if(lookup_socket(oldsock)->flags & CEPT_UDP_USED) MOD_INC_USE_COUNT;
   //MOD_INC_USE_COUNT; //UDP

  sock =  (getsockop(oldsock))->dup(newsock, oldsock);
  csp = add_socket(newsock);
  csp->flags = lookup_socket(oldsock)->flags;
  return sock;
}

/*
 * Intercept release calls.  Free up any memory used and decrement use count.
 */
static int udp_release(struct socket *sock, struct socket *peer) {
#if DEBUG>1
   printk("<1>IPcept (udp_release): releasing socket %x\n",sock);
#endif
   if (sock->type == SOCK_DGRAM) {
     struct cept *csp;
#if DEBUG>4
     printk("<4>IPcept (udp_release): looking up cept socket\n");
#endif
       csp = lookup_socket(sock);
       
       if (csp && HAS_ALT_REMOTE(csp)) {
	 int i;
#if DEBUG>4
	 printk("<4>IPcept (udp_release): cept has proxy\n");
#endif
	 
           for (i = 0; i < csp->num_close; i++) {
               /*
                * Send a close message to the proxy; close message
                * is a message that is sent to the proxy port with
                * original destination address of INADDR_ANY port 0
                */
               struct cept *proxy;
               struct msghdr m;
	       struct iovec iov;
               struct sockaddr_in close_addr, proxy_addr;
               int total_len, nonblock, flags;
	       struct scm_cookie *scm;
	       m.msg_controllen = 0;
	       m.msg_flags = flags | MSG_DONTWAIT;
               iov.iov_base = (void *) &close_addr;
               iov.iov_len = total_len;
               
              (getsockop(sock))->sendmsg(sock, &m, total_len, scm);
           }
       }
     }
   /* in case this socket was managed, delete it */

#if DEBUG>4
	 printk("<4>IPcept (udp_release): decrementing use count\n");
#endif
   if(lookup_socket(sock)!=NULL) {
     if(lookup_socket(sock)->flags & CEPT_UDP_USED)  MOD_DEC_USE_COUNT;
     //else printk("<4>IPcept (udp_release): unmanaged cept?\n");
     }
   //else printk("<4>IPcept (udp_release): cept is null\n");
   
#if DEBUG>4
	 printk("<4>IPcept (udp_release): deleting proxy\n");
#endif
   del_cept_socket(sock);

#if DEBUG>4
	 printk("<4>IPcept (udp_release): returning original release\n");
#endif
   //MOD_DEC_USE_COUNT; //UDP
   return (getsockop(sock))->release(sock, peer);
}

static int udp_bind(struct socket *sock, struct sockaddr *addr, 
                     int sockaddr_len) {
   struct sockaddr_in *sin=(struct sockaddr_in *)addr;

   if (sockaddr_len < sizeof(struct sockaddr_in)) {
      return -EINVAL;
   }

   if ((sin->sin_addr.s_addr == INADDR_ANY) || 
       (inet_addr_type(sin->sin_addr.s_addr) == RTN_LOCAL)) {
      return (getsockop(sock))->bind(sock, addr, sockaddr_len);
   } else {
      struct cept *csp = lookup_socket(sock);

      /* must be superuser to masquerade */
      if (! oldsuser()) {
         return -EPERM;
      }

      if (csp == NULL) {
         csp = add_socket(sock);
      }

#if DEBUG>0
   printk("<1>IPcept (udp_bind): binding socket %x to foreign address\n",sock);
#endif

      SET_HAS_ALT_LOCAL(csp);
      memcpy(&(csp->local_addr), addr, sockaddr_len);
   }
}

/*
 * Intercept connect calls.  If there is a proxy for our dest port,
 * change the destination address to point to the proxy.  If not,
 * and we're not managed for some other reason (perhaps masquerading),
 * we no longer need to manage this socket.
 */
static int udp_connect(struct socket *sock, struct sockaddr *servaddr,
                        int sockaddr_len, int flags) {
   struct cept *csp = lookup_socket(sock);

   if (sockaddr_len < sizeof(struct sockaddr_in)) {
      return -EINVAL;
   }
   
   if (sock->type == SOCK_DGRAM) {
     struct sockaddr_in *sin = (struct sockaddr_in *) servaddr;
     
     if (csp == NULL)
       csp = add_socket(sock);
     
       csp->udp_connected = 1;
#if DEBUG>1
     printk("<1>IPcept (cept_connect): UDP socket %x doing a connect\n", sock);
#endif
     csp->udp_conn_addr.sin_addr.s_addr = sin->sin_addr.s_addr;
     csp->udp_conn_addr.sin_port = sin->sin_port;
     
     return (getsockop(sock))->connect(sock, servaddr, sockaddr_len, flags);
   }
   if ( (! csp) || (! DONT_PROXY(csp)) ) {
     struct sockaddr_in *sin=(struct sockaddr_in *)servaddr;
     struct cept *proxy = lookup_proxy(sock->type, ntohs(sin->sin_port));
     
     if ((proxy!=NULL) && (proxy->sk->num != 0)) {
       
       if (csp == NULL) {
	 csp = add_socket(sock);
       }
       
       /* save the originally requested address */
       SET_HAS_ALT_REMOTE(csp);
       memcpy(&(csp->remote_addr), sin, sizeof(struct sockaddr_in));
       
       /* if SOCKADDR_ANY was requested, get address from proxy socket */
       if (csp->remote_addr.sin_addr.s_addr == 0) {
	 csp->remote_addr.sin_addr.s_addr = proxy->sk->saddr;
       }
       
       /* replace address with a local address */
       sin->sin_addr.s_addr = INADDR_ANY;   /* local */
       sin->sin_port = htons(proxy->sk->num);
       
#if DEBUG>0
       printk("<1>IPcept (cept_connect): redirecting socket %x type %d ",
	      sock, sock->type);
       printk("to %d %d\n", ntohl(sin->sin_addr.s_addr), ntohs(sin->sin_port));
       
       
#endif
     }
   }
 
   /*
    * Unless there's some other reason we're managing this socket,
    * we can unmanage it.
    */
   if (! (csp && NEEDS_MANAGEMENT(csp)) && (sock->type == SOCK_STREAM)) {
      /* Port not intercepted; use normal operations */
      if (csp) {
         del_cept_socket(sock);
      }
      unmanage(sock);
   }

   return (getsockop(sock))->connect(sock, servaddr, sockaddr_len, flags);
}

static int udp_socketpair(struct socket *sock1, struct socket *sock2) {
   return (getsockop(sock1))->socketpair(sock1, sock2);
}

/*
 * Intercept the accept call.  For the new socket, look for the peer socket 
 * and copy any alternate addresses.  If there's nothing special about this 
 * socket by now, there never will be, unmanage it.
 */
static int udp_accept(struct socket *sock, struct socket *newsock, 
                       int flags) {
   int res = (getsockop(sock))->accept(sock, newsock, flags);
   struct sock *sk=(struct sock *)newsock->sk;
   struct cept *csp = NULL;
   struct cept *peer;

   if (res<0) {
      unmanage(newsock);
      return res;
   }

   peer = lookup_address(ntohl(sk->daddr), ntohs(sk->dport));
   if (peer != NULL) {
      csp = lookup_socket(newsock);
      if (csp == NULL) {
         csp = add_socket(newsock);
      }

      /* is the client masquerading? */
      if (HAS_ALT_LOCAL(peer)) {
         SET_HAS_ALT_REMOTE(csp);
         memcpy(&(csp->remote_addr), &(peer->local_addr), 
                                        sizeof(struct sockaddr_in));
#if DEBUG>1
         printk("<1>IPcept (acudp): socket %x accepted connection from masquerading peer\n", sock);
#endif
      }

      /* was the client proxied? */
      if (HAS_ALT_REMOTE(peer)) {
         SET_HAS_ALT_LOCAL(csp);
         memcpy(&(csp->local_addr), &(peer->remote_addr), 
                                        sizeof(struct sockaddr_in));
#if DEBUG>1
         printk("<1>IPcept (accept): socket %x accepted proxied connection\n", sock);
#endif
      }
   }

   /* If we're not managing the new sock now, we no longer need to */
   if (! (csp && NEEDS_MANAGEMENT(csp)) ) {
      if (csp) {
         del_cept_socket(newsock);
      }
      unmanage(newsock);
   }

   return res;
}

/* 
 * If an alternate name of the type requested (peer or not) is requested,
 * use it.  Otherwise, fall through to the normal operation.  Note, this
 * function is also called during the return from accept().
 */
static int udp_getname(struct socket *sock, struct sockaddr *addr,
                        int *sockaddr_len, int peer) {
   struct cept *cp = lookup_socket(sock);

   if ((peer) && (cp != NULL) && HAS_ALT_REMOTE(cp)) {
      *sockaddr_len = sizeof(cp->remote_addr);
      memcpy(addr, &(cp->remote_addr), *sockaddr_len);
      return 0;
   } else if ((! peer) && (cp != NULL) && HAS_ALT_LOCAL(cp)) {
      *sockaddr_len = sizeof(cp->local_addr);
      memcpy(addr, &(cp->local_addr), *sockaddr_len);
      return 0;
   } else {
      return (getsockop(sock))->getname(sock, addr, sockaddr_len, peer);
   }
}

static unsigned int udp_poll(struct file *file, struct socket *sock, poll_table *wait) {
   return (getsockop(sock))->poll(file, sock, wait);
}

static int udp_ioctl(struct socket *sock, unsigned int cmd, 
                      unsigned long arg) {
   return (getsockop(sock))->ioctl(sock, cmd, arg);
}

static int udp_listen(struct socket *sock, int len) {
   return (getsockop(sock))->listen(sock, len);
}

static int udp_shutdown(struct socket *sock, int flags) {
   return (getsockop(sock))->shutdown(sock, flags);
}




/* 
 * Handle SOL_IPCEPT level calls or pass the call along to IP.  Most
 * calls require superuser privlidges.  This call frequently results
 * in the creation of a new cept struct to hold the info.
 */

static int udp_setsockopt(struct socket *sock, int level, int optname,
                           char *optval, int optlen) {
   struct sock *sk=(struct sock *)sock->sk;
   int i;
#if DEBUG>1
   printk("<1>IPcept (udp_setsockopt): called for %x with %d, %d\n", 
          sock, level, optname);
#endif

   if (level == SOL_IPCEPT) {
      int err=0;
      struct cept *csp = lookup_socket(sock);
      struct cept *prox;
      /* must be superuser to set most options */
      
      printk("<1> IPcept: %d,%d,%d\n",current->uid,current->euid,current->fsuid);
      if (PRIVILIGED_OPTION(optname) && (!oldsuser())) {
	
	printk("<1>IPcept (udp_setsockopt): not superuser\n");
         return -EPERM;
      }


      if (csp == NULL) {
         csp = add_socket(sock);
      }

      if (optname == IPCEPT_INTERCEPT) {
         int val;

         if (optlen < sizeof(int)) {
            err = -EINVAL;
         }

         if (!err) {
            err=verify_area(VERIFY_READ, optval, sizeof(int));
         }

         if(!err) {
	   //val = get_user((int *) optval);
	   get_user(val, (int *) optval); //returns error code, 0 for success
           printk("udp_setsockopt: val set to %d\n",val);
	   /* Is there already a proxy for this port? */
	   if (prox = lookup_proxy(sock->type, ntohs(val))) {         
#if DEBUG>2
	     printk("<1>%s: Socket %x already proxies %d\n",
                    "IPcept (setsockopt)", prox->sock, ntohs(val));
#endif	
	     err = -EADDRINUSE;
	     /* fall through for error handling */
	   } else {
	     /* proxy this port */
               SET_IS_PROXY(csp);
               if (csp->num_ports >= MAX_PROXY_PORTS) {
                  return -EPERM;
               }
               csp->redir_port[(csp->num_ports)++] = ntohs(val);
#if DEBUG>0
               printk("<1>%s: Socket %x proxies port %d to port %d\n",
                      "IPcept (setsockopt)", sock, ntohs(val), sk->num);
#endif
	       lookup_socket(sock)->flags |= CEPT_UDP_USED;
	       //if(lookup_socket(sock)->flags & CEPT_UDP_USED) MOD_INC_USE_COUNT;
	       MOD_INC_USE_COUNT;
	       return 0;  /* success! */
            }
         }
         /* on error, fall through to error handling code */
  } else if (optname == IPCEPT_DONT_INTERCEPT) {
          /* Remove this port number from the ports we intercept
             Should always return 0 (even when we don't currently
             intercept on that port, unless there is some major error */

          int val;
          
          if (optlen < sizeof(int)) {
              err = -EINVAL;
          }

          if (!err) {
              err = verify_area(VERIFY_READ, optval, sizeof(int));
          }

          if(!err) {
              struct cept *proxy;
              get_user(val,(int *) optval);
	      val = ntohs(val);

              /* Is there already a proxy for this port? */
              if ((proxy = lookup_proxy(sock->type, val)) != NULL) {
                  if (proxy->sock == sock) {
                      int i = 0;

#if DEBUG>0
                      printk("<1>IPcept (cept_setsockopt): masquerading disabled on sock %x for port %d\n", sock, val);
#endif
                      
                      /* Know it's in there somewhere before proxy->num_ports */
                      while (proxy->redir_port[i] != val)
                          i++;

                      proxy->num_ports--;

                      /* If it's not the last one, just replace the deleted
                         proxy num with the last number */
                      if (i != proxy->num_ports)
                          proxy->redir_port[i] = proxy->redir_port[proxy->num_ports];
                  }
              }

              return 0;
	    }
          

      } else if (optname == IPCEPT_MASQUERADE) {
	if (sock->type == SOCK_DGRAM) {
	  struct sockaddr_in masq_addr;
	  if (optlen < (2 * sizeof(struct sockaddr_in)))
	    err = -EINVAL;
	  
	  if (!err)
	    err = verify_area(VERIFY_READ, optval,
			      2 * sizeof(struct sockaddr_in));
	  if(err) printk("<4>IPcept (udp_setsockopt): error in verifying area.");
	  if (!err) {
                  /*
                   * Note on formats:
                   * sockopt optval format is 2 struct sockaddr_in:
                   *     1st - Dest Addr
                   *     2nd - Src Addr (who to masquerade as)
                   *
                   * Internal format is 1 struct sockaddr_in
                   *     addr = Dest Addr
                   *     (struct sockaddr_in *) addr.sin_zero = Src Addr
                   */
                  copy_from_user(&masq_addr, optval,  sizeof(struct sockaddr_in));
                  /* Now copy the src addr we want to masquerade as, but
                     don't include sin_zero (size char[8]) */

		  copy_from_user(masq_addr.sin_zero,
				 optval + sizeof(struct sockaddr_in),
				 sizeof(struct sockaddr_in) - sizeof(char[8]));
		  printk("<4>IPcept : %d", masq_addr);
		  
                  SET_HAS_ALT_LOCAL(csp);
#if DEBUG>0
                  printk("<1>IPcept (cept_setsockopt): masquerading enabled on %x\n", sock);
#endif
                  memcpy(&(csp->local_addr), &masq_addr,
                      sizeof(csp->local_addr));
		  //if(lookup_socket(sock)->flags & CEPT_UDP_USED) MOD_INC_USE_COUNT;                     
		  lookup_socket(sock)->flags |= CEPT_UDP_USED;
		  MOD_INC_USE_COUNT;
                  return 0;
		}
              /* on error, fall through to error handling code */
	    }
          


   } else if (optname == IPCEPT_DONT_PROXY) {

#if DEBUG>0
   printk("<1>IPcept (cept_setsockopt): proxying disabled on %x\n",sock);
#endif

         SET_DONT_PROXY(csp);
         return 0;

      } else if (optname == IPCEPT_DONT_MASQUERADE) {


   if (sock->type == SOCK_STREAM) {
#if DEBUG>0
              printk("<1>IPcept (cept_setsockopt): masquerading disabled on %x\n",sock);
#endif
              SET_DONT_MASQUERADE(csp);
              return 0;
          } else if (sock->type == SOCK_DGRAM) {
              struct sockaddr_in addr;

              if (optlen < sizeof(struct sockaddr_in))
                  err = -EINVAL;

              if (!err)
                  err = verify_area(VERIFY_READ, optval, sizeof(struct sockaddr_in));

              if (!err) {
                  copy_from_user(&addr, optval, sizeof(struct sockaddr_in));
              }

#if DEBUG>0
              printk("<1>IPcept (cept_setsockopt): masquerading disabled on %x\n", sock);
#endif
                  csp->flags &= ~(CEPT_ALT_LOCAL);
                  return 0;
          }


      } else if (optname == IPCEPT_UNMANAGE) {

         del_cept_socket(sock);
         unmanage(sock);
         return 0;

      } else {
         err = -ENOPROTOOPT;
         /* fall through to error handling code */
      }

      /* errors fall through to here */
      if (!NEEDS_MANAGEMENT(csp)) {
         del_cept_socket(sock);
      }
      return err;
   } else {
      return (getsockop(sock))->setsockopt(sock, level, optname, optval, optlen);
   }
}


static int udp_getsockopt(struct socket *sock, int level, int optname,
                           char *optval, int *optlen) {
   if ((level == SOL_SOCKET) && (optname == SO_BINDTODEVICE)) {
      struct cept *csp = lookup_socket(sock);

      if ((csp!=NULL) && (HAS_ALT_LOCAL(csp))) {
         struct ifreq req;
	 int err=verify_area(VERIFY_WRITE,optval,sizeof(req));
	 if(err<0) return err;
	 
         if (csp->sk->bound_dev_if && dev_get_by_index(csp->sk->bound_dev_if)) {
            strncpy(req.ifr_name, (dev_get_by_index(csp->sk->bound_dev_if))->name, 
                                           sizeof(req.ifr_name));
            (*(struct sockaddr_in *) &req.ifr_addr).sin_family = csp->sk->family;
            (*(struct sockaddr_in *) &req.ifr_addr).sin_addr.s_addr = 
                                               csp->local_addr.sin_addr.s_addr;
         }

	 if(copy_to_user(optval, &req, sizeof(req))) return -EFAULT;

         return 0;
      }
     } else if (level == SOL_IPCEPT) {
       int err;
       struct cept *csp = lookup_socket(sock);
       
       if ((optname == IPCEPT_LAST_RECV_DEST) && (sock->type == SOCK_DGRAM)) {
           int new_optlen;
           if ((csp == NULL) || !IS_PROXY(csp)) {
	     printk("<4> IPcept (udp_getsockopt): csp null or not a proxy.\n"); 
	     return -EINVAL;
	   }

           err=verify_area(VERIFY_READ, optlen, sizeof(int));
           if (err < 0)
               return err;

           copy_from_user(&new_optlen, optlen, sizeof(int));
           if (new_optlen < sizeof(struct sockaddr_in))
               return -EINVAL;
                      
           err = verify_area(VERIFY_WRITE, optval, sizeof(struct sockaddr_in));
           if (err < 0)
               return err;

           copy_to_user(optval, &(csp->last_recv_dest), sizeof(struct sockaddr_in));
           new_optlen = sizeof(struct sockaddr_in);
           copy_to_user(optlen, &new_optlen, sizeof(int));
           return 0;
           
       } else {
           err = -ENOPROTOOPT;
       }

       return err;
   }
     
   return (getsockop(sock))->getsockopt(sock, level, optname, optval, optlen);
}

static int udp_fcntl(struct socket *sock, unsigned int cmd, 
                      unsigned long arg) {
   return (getsockop(sock))->fcntl(sock, cmd, arg);
}

static int udp_sendmsg(struct socket *sock, struct msghdr *m, int total_len,          struct scm_cookie *scm) {

  struct sockaddr_in tmp_dest_addr;
  int len;
  
    if (sock->type == SOCK_DGRAM) {
        struct cept *csp = lookup_socket(sock);
        struct sockaddr_in *dest_addr = (struct sockaddr_in *) m->msg_name;
        struct cept *proxy = NULL;
        
        /* If they did a connect, get the real dest address here */
        if (csp && csp->udp_connected) {
#if DEBUG>1
            printk("<1>IPcept (cept_sendmsg): UDP socket %x did a connect, getting real dest address\n", sock);
#endif
            m->msg_name = dest_addr = &tmp_dest_addr;
            m->msg_namelen = sizeof(tmp_dest_addr);
            dest_addr->sin_family = AF_INET;
            dest_addr->sin_addr.s_addr = csp->udp_conn_addr.sin_addr.s_addr;
            dest_addr->sin_port = csp->udp_conn_addr.sin_port;
        }
            
        if (!(csp && DONT_PROXY(csp)))
            proxy = lookup_proxy(SOCK_DGRAM, ntohs(dest_addr->sin_port));

        if (csp && HAS_ALT_LOCAL(csp)) {
#if DEBUG>4
            printk("<1>IPcept (cept_sendmsg): masquerding on send, UDP socket %x\n", sock);
#endif
            memcpy(dest_addr, &(csp->local_addr), sizeof(struct sockaddr_in));
            m->msg_flags |= MSG_PROXY;
        }

        if ((proxy != NULL) && (proxy->sk->num != 0)) {
            unsigned char *buf;
            mm_segment_t fs;
            int err;

            if ((total_len + sizeof(struct sockaddr_in)) > 65515)
                return -EMSGSIZE;
	    len = total_len + sizeof(struct sockaddr_in);
            buf = kmalloc(len, GFP_KERNEL);
            if (buf == NULL) {
#if DEBUG>1
                printk("<1>IPcept (cept_sendmsg): kmalloc couldn't allocate memory\n");
#endif
                return -ENOBUFS;
            }


#if DEBUG>3
            printk("<1>IPcept (cept_sendmsg) socket %x proxying datagram from port ", sock);
            printk("%d to localhost:%d\n", ntohs(dest_addr->sin_port),
                proxy->sk->num);
#endif

            add_close_port(sock, ntohs(dest_addr->sin_port));

            memcpy_fromiovec(buf, m->msg_iov, total_len);
            fs = get_fs();
            set_fs(get_ds());

            memcpy(buf + total_len, dest_addr, sizeof(struct sockaddr_in));

            total_len += sizeof(struct sockaddr_in);
            m->msg_iovlen = 1;
            m->msg_iov[0].iov_base = buf;
            m->msg_iov[0].iov_len = total_len;
                
            /* Change Destination Address to local */
            dest_addr->sin_addr.s_addr = INADDR_ANY;
            dest_addr->sin_port = htons(proxy->sk->num);

            /* Do it */
            err = (getsockop(sock))->sendmsg(sock, m, total_len, scm);

            set_fs(fs);
            kfree_s(buf, len);

            if (err > 0) {
                /* Account for change of size in datagram */
                return (err - sizeof(struct sockaddr_in));
            } else {
                printk("<1>IPcept (cept_sendmsg): error %d on send, UDP socket %x\n",
                    err, sock);
                return err;
            }
        }
    }
   return (getsockop(sock))->sendmsg(sock, m, total_len, scm);
}

static int udp_recvmsg(struct socket *sock, struct msghdr *m, int total_len, 
                        int flags, struct scm_cookie *scm) {

  struct cept *csp = lookup_socket(sock);
    
    if ((sock->type == SOCK_DGRAM) && csp && IS_PROXY(csp)) {
        int err;
	mm_segment_t fs;
        char *buf, *old_buf;
        struct sock *sk = (struct sock *) sock->sk;

        buf = kmalloc(m->msg_iov[0].iov_len, GFP_KERNEL);
        if (buf == NULL) {
#if DEBUG>1
            printk("<1>IPcept (cept_recvmsg): kmalloc couldn't allocate memory\n");
#endif
            return -ENOBUFS;
        }

        old_buf = m->msg_iov[0].iov_base;
        m->msg_iov[0].iov_base = buf;

        fs = get_fs();
        set_fs(get_ds());
        err = (getsockop(sock))->recvmsg(sock, m, total_len, flags, scm);
        set_fs(fs);
        
        if (err > 0) {
            /* Fix to check if the message was proxied or not.
               Non-proxied src addresses are set using sockopt???
               For now, blindly assume it was proxied */
            int read_len = err - sizeof(struct sockaddr_in);
            
            memcpy(&(csp->last_recv_dest), buf + read_len,
                sizeof(struct sockaddr_in));
            copy_to_user(old_buf, buf, read_len);
            
            err = read_len;

#if DEBUG>4
            printk("<1>IPcept (cept_recvmsg): receiving proxied message\n");
#endif
            kfree(buf);
        }
        
        m->msg_iov[0].iov_base = old_buf;
        return err;
      }

   return (getsockop(sock))->recvmsg(sock, m, total_len, flags, scm);
}


/*

/***********************************************************************
 **                   Procfs access                                   **
 ***********************************************************************/

/*
 * Handle reads to our status file in the procfs.
 */
static int ipcept_read(char *buf, char **start, off_t offset, int len, 
                       int unused) {
   int limit = PAGE_SIZE - 160;
   struct cept *cp;

   len=0;

   for (cp = cept_socks; cp != NULL; cp = cp->next) {
#if DEBUG>0
      len += sprintf(buf+len,"%x:  ", cp->sock);
#endif

      len += sprintAddress(buf+len, ntohl(cp->sk->saddr));
      len += sprintf(buf+len,":%d ", cp->sk->num);
      if (IS_PROXY(cp)) {
         int i;
         len += sprintf(buf+len, "[");
         for (i=0; i<cp->num_ports; i++) {
            len += sprintf(buf+len, "%d", cp->redir_port[i]);
            if ((i+1) < cp->num_ports) {
               len += sprintf(buf+len, ",");
            }
         }
         len += sprintf(buf+len, "]");
      }
      if (HAS_ALT_LOCAL(cp)) {
         len += sprintf(buf+len, "(");
         len += sprintAddress(buf+len, ntohl(cp->local_addr.sin_addr.s_addr));
         len += sprintf(buf+len, ":%d)	", ntohs(cp->local_addr.sin_port));
      } else {
         len += sprintf(buf+len, "		");
      }
      len += sprintf(buf+len, "-> ");
      len += sprintAddress(buf+len, ntohl(cp->sk->daddr));
      len += sprintf(buf+len, ":%d ", ntohs(cp->sk->dport));
      if (HAS_ALT_REMOTE(cp)) {
         len += sprintf(buf+len, "(");
         len += sprintAddress(buf+len, ntohl(cp->remote_addr.sin_addr.s_addr));
         len += sprintf(buf+len, ":%d) ", ntohs(cp->remote_addr.sin_port));
      }
      len += sprintf(buf+len,"\n");
      if (len>limit) {
         return len;
      }
   }
   return len;
}


/***********************************************************************
 **                   Module setup and cleanup                        **
 ***********************************************************************/

/*
 * Setup the newly loaded module.
 */
int init_module(void) {
#if DEBUG>0
   printk("<1>IPcept: starting up\n");
#endif

   cept_socks = NULL;

   /* Replace the IP operations with our own.  Get back the old IP ops */
   ipfamily = (struct net_proto_family *) sock_replace(IPcept_proto_family.family, &IPcept_proto_family);

#if DEBUG>0
   printk("<1>IPcept: replaced ip protocol family\n");
#endif

   /* Register our status file in procfs (under net) */
   proc_net_register(&ipcept_proc_entry);

   return 0;
}

/*
 * Cleanup the module for shutdown.
 */
void cleanup_module(void) {
   /* put the old IP operations back */
   sock_unregister(PF_INET);
   sock_register(ipfamily);

#if DEBUG>0
   printk("<1>IPcept: restored ip protocol family\n");
#endif

   /* Remove any remaining allocated structures (there shouldn't be any) */
   cleanup_cept_socks();

   /* Unregister our procfs file */
   proc_net_unregister(ipcept_proc_entry.low_ino);

#if DEBUG>0
   printk("<1>IPcept: shutting down\n");
#endif
}

asmlinkage int oldsuser(void)
{
        if (current->uid == 0) {
                current->flags |= PF_SUPERPRIV;
                return 1;
        }
        return 0;
}



