/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define SOL_IPCEPT		99
#define IPCEPT_INTERCEPT	1
#define IPCEPT_MASQUERADE	2
#define IPCEPT_DONT_PROXY	3
#define IPCEPT_DONT_MASQUERADE	4
#define IPCEPT_UNMANAGE		5
#define IPCEPT_LAST_RECV_DEST   6
#define IPCEPT_DONT_INTERCEPT   7

#define MAX_PROXY_PORTS		10  /* number of ports a socket can proxy */

#ifdef __KERNEL__

#define UNPRIVILIGED_OPTION(optname) ((optname == IPCEPT_DONT_PROXY) || \
                                      (optname == IPCEPT_DONT_MASQUERADE) || \
                                      (optname == IPCEPT_UNMANAGE))
#define PRIVILIGED_OPTION(optname) (! UNPRIVILIGED_OPTION(optname))

#define CEPT_PROXY		0x1
#define CEPT_ALT_LOCAL		0x2
#define CEPT_ALT_REMOTE		0x4
#define CEPT_DONT_PROXY		0x8
#define CEPT_DONT_MASQUERADE	0x10
#define CEPT_UDP_USED           0x20
#define IS_PROXY(cp)			((cp)->flags & CEPT_PROXY)
#define SET_IS_PROXY(cp)		(cp)->flags |= CEPT_PROXY

#define HAS_ALT_LOCAL(cp)		((cp)->flags & CEPT_ALT_LOCAL)
#define SET_HAS_ALT_LOCAL(cp)		(cp)->flags |= CEPT_ALT_LOCAL

#define HAS_ALT_REMOTE(cp)		((cp)->flags & CEPT_ALT_REMOTE)
#define SET_HAS_ALT_REMOTE(cp)		(cp)->flags |= CEPT_ALT_REMOTE

#define DONT_PROXY(cp)			((cp)->flags & CEPT_DONT_PROXY)
#define SET_DONT_PROXY(cp)		(cp)->flags |= CEPT_DONT_PROXY

#define DONT_MASQUERADE(cp)		((cp)->flags & CEPT_DONT_MASQUERADE)
#define SET_DONT_MASQUERADE(cp)		(cp)->flags |= CEPT_DONT_MASQUERADE

#define NEEDS_MANAGEMENT(cp)		(cp->flags != 0)

struct cept {
   int flags;
   struct sockaddr_in local_addr;  /* other local address (net byte order) */
   struct sockaddr_in remote_addr; /* other remote address (net byte order) */
   struct sockaddr_in udp_conn_addr; /* UDP connection address */
   struct sockaddr_in last_recv_dest; /* Destination address for last received
					 message (UDP only, net byte order */
   int num_ports;                  /* number of redirected ports */
   unsigned short redir_port[MAX_PROXY_PORTS];
                                   /* proxy port (in host byte order) */
   int udp_connected;
   int num_close;
   unsigned short close_port[MAX_PROXY_PORTS]; /* which port numbers we sent 
						  to that were proxied, host 
						  byte order.  This keeps 
						  track of who we need to 
						  send close notification to. */
   struct socket *sock;            /* a pointer to the socket */
   struct sock *sk;
   struct cept *next;
  };


/******************************************************************************
 * IPcept Prototypes
 *****************************************************************************/

static int del_cept_socket(struct socket *sock);
static struct cept *add_socket(struct socket *sock);
static struct cept *lookup_proxy(int type, unsigned short redir);
static struct cept *lookup_socket(struct socket *sock);
static struct cept *lookup_address(long addr, unsigned short port);
int add_close_port (struct socket *sock, unsigned short num);
static void cleanup_cept_socks();
static void unmanage(struct socket * sock);
static int sprintAddress(char *buf, unsigned long addr);
static int cept_create(struct socket *sock, int protocol);
static int cept_dup(struct socket *newsock, struct socket *oldsock);
static int cept_release(struct socket *sock, struct socket *peer);
static int cept_bind(struct socket *sock, struct sockaddr *addr, 
		     int sockaddr_len);
static int cept_connect(struct socket *sock, struct sockaddr *servaddr, 
			int sockaddr_len, int flags);
static int cept_socketpair(struct socket *sock1, struct socket *sock2);
static int cept_accept(struct socket *sock, struct socket *newsock, int flags);
static int cept_getname(struct socket *sock, struct sockaddr *addr,   
			int *sockaddr_len, int peer);
static unsigned int cept_poll(struct file *file, struct socket *sock, poll_table *wait);
static int cept_ioctl(struct socket *sock, unsigned int cmd, 
		      unsigned long arg);
static int cept_listen(struct socket *sock, int len);
static int cept_shutdown(struct socket *sock, int flags);
static int cept_setsockopt(struct socket *sock, int level, int optname,
                           char *optval, int optlen);
static int cept_getsockopt(struct socket *sock, int level, int optname,
                           char *optval, int *optlen);
static int cept_fcntl(struct socket *sock, unsigned int cmd, 
                      unsigned long arg);
static int cept_sendmsg(struct socket *sock, struct msghdr *msg, int size, 
                        struct scm_cookie *scm);
static int cept_recvmsg(struct socket *sock, struct msghdr *m, int total_len, 
                        int flags, struct scm_cookie *scm);

static int udp_create(struct socket *sock, int protocol);
static int udp_dup(struct socket *newsock, struct socket *oldsock);
static int udp_release(struct socket *sock, struct socket *peer);
static int udp_bind(struct socket *sock, struct sockaddr *addr, 
		     int sockaddr_len);
static int udp_connect(struct socket *sock, struct sockaddr *servaddr, 
			int sockaddr_len, int flags);
static int udp_socketpair(struct socket *sock1, struct socket *sock2);
static int udp_accept(struct socket *sock, struct socket *newsock, int flags);
static int udp_getname(struct socket *sock, struct sockaddr *addr,   
			int *sockaddr_len, int peer);
static unsigned int udp_poll(struct file *file, struct socket *sock, poll_table *wait);
static int udp_ioctl(struct socket *sock, unsigned int cmd, 
		      unsigned long arg);
static int udp_listen(struct socket *sock, int len);
static int udp_shutdown(struct socket *sock, int flags);
static int udp_setsockopt(struct socket *sock, int level, int optname,
                           char *optval, int optlen);
static int udp_getsockopt(struct socket *sock, int level, int optname,
                           char *optval, int *optlen);
static int udp_fcntl(struct socket *sock, unsigned int cmd, 
                      unsigned long arg);
static int udp_sendmsg(struct socket *sock, struct msghdr *msg, int size, 
                        struct scm_cookie *scm);
static int udp_recvmsg(struct socket *sock, struct msghdr *m, int total_len, 
                        int flags, struct scm_cookie *scm);

static int ipcept_read(char *buf, char **start, off_t offset, int len, 
                       int unused);
  
/******************************************************************************
 * Procfs dir entry 
 *****************************************************************************/

struct proc_dir_entry ipcept_proc_entry = {
        0,			/* inode */
	6, "IPcept",		/* name len and name */
	S_IFREG | S_IRUGO, 	/* perms */
	1, 0, 0,		/* nliks, owner, group */
	0,			/* size (unused) */
	NULL,			/* ops */
	&ipcept_read		/* read func */
};

int init_module(void);
void cleanup_module(void);

/******************************************************************************
 * Structure to replace the normal IP operations structure with our own.
 *****************************************************************************/

static struct proto_ops IPcept_tcp_ops = {
	PF_INET,

        cept_dup,
        cept_release,   
        cept_bind,
        cept_connect,
        cept_socketpair,
        cept_accept,
        cept_getname,
        cept_poll,    
        cept_ioctl,     
        cept_listen,
        cept_shutdown,
        cept_setsockopt,
        cept_getsockopt,
        cept_fcntl,
        cept_sendmsg,
        cept_recvmsg
};

static struct proto_ops IPcept_udp_ops = {
	PF_INET,

        udp_dup,
        udp_release,   
        udp_bind,
        udp_connect,
        udp_socketpair,
        udp_accept,
        udp_getname,
        udp_poll,    
        udp_ioctl,     
        udp_listen,
        udp_shutdown,
        udp_setsockopt,
        udp_getsockopt,
        udp_fcntl,
        udp_sendmsg,
        udp_recvmsg
};








static struct net_proto_family IPcept_proto_family = {
        PF_INET,
	cept_create
};



#endif


