/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#include "util.h"
#include "../IPcept.h"

#include "FMG_IPcept_SocketControl.h"

JNIEXPORT void JNICALL Java_FMG_IPcept_SocketControl_shutdownInputStream
  (JNIEnv *env, jclass thisClass, jobject is) {
   jclass isclass = env->GetObjectClass(is);
   jfieldID filedesID = env->GetFieldID(isclass, "fd", 
                                        "Ljava/io/FileDescriptor;");
   jobject fdesobj = env->GetObjectField(is, filedesID);

   int fd = getFDfromFileDes(env, fdesobj);

   if (shutdown(fd, 0)<0) {
      jclass excl = env->FindClass("java/io/IOException");
      env->ThrowNew(excl, "shutdown() failed");
      return;
   }
}

JNIEXPORT void JNICALL Java_FMG_IPcept_SocketControl_shutdownOutputStream
  (JNIEnv *env, jclass thisClass, jobject os) {
   jclass osclass = env->GetObjectClass(os);
   jfieldID filedesID = env->GetFieldID(osclass, "fd", 
                                        "Ljava/io/FileDescriptor;");
   jobject fdesobj = env->GetObjectField(os, filedesID);

   int fd = getFDfromFileDes(env, fdesobj);

   if (shutdown(fd, 1)<0) {
      jclass excl = env->FindClass("java/io/IOException");
      env->ThrowNew(excl, "shutdown() failed");
      return;
   }
}

JNIEXPORT void JNICALL Java_FMG_IPcept_SocketControl_shutdownSock
  (JNIEnv *env, jclass thisClass, jobject socket, jint mode) {

   int fd = getFDfromSocket(env, socket);

   if (shutdown(fd, mode)<0) {
      jclass excl = env->FindClass("java/io/IOException");
      env->ThrowNew(excl, "shutdown() failed");
      return;
   }
}

JNIEXPORT jobject JNICALL Java_FMG_IPcept_SocketControl_createUnproxiedSocket
  (JNIEnv *env, jclass thisClass, jobject destAddr, jint destPort, jobject srcAddr, jint srcPort) {
   /* create a new impl */
   jclass implclass = env->FindClass("java/net/PlainSocketImpl");
   jmethodID implinit = env->GetMethodID(implclass, "<init>", "()V");
   jobject impl = env->NewObject(implclass, implinit);

   /* call impl.create(1) */
   jmethodID create = env->GetMethodID(implclass, "create", "(Z)V");
   env->CallVoidMethod(impl, create, JNI_TRUE);

   /* get the fd */
   int fd = getFDfromSocketImpl(env, impl);

   /* set the no_proxy option */
   if (setsockopt(fd, SOL_IPCEPT, IPCEPT_DONT_PROXY, 0, 0)<0) {
      jclass excl = env->FindClass("java/io/IOException");
      env->ThrowNew(excl, "setsockopt() failed");
      return NULL;
   }

   /* call bind if needed */
   if (srcAddr!=NULL) {
      jmethodID bind = env->GetMethodID(implclass, "bind", 
                                        "(Ljava/net/InetAddress;I)V");
      env->CallVoidMethod(impl, bind, srcAddr, srcPort);

      /* if the find throws an exception, report it to the user */
      if (env->ExceptionOccurred()) {
         return NULL;
      }
   }

   /* call connect */
   jmethodID connect = env->GetMethodID(implclass, "connect", 
                                        "(Ljava/net/InetAddress;I)V");
   env->CallVoidMethod(impl, connect, destAddr, destPort);

   /* Catch SocketException, call impl.close() and re-throw */
   jthrowable except = env->ExceptionOccurred();
   if (except != NULL) {
      jmethodID close = env->GetMethodID(implclass, "close", "()V");
      env->CallVoidMethod(impl, close);
      env->Throw(except);
      return NULL;
   }

   /* create a socket with this implementation, and return */
   jclass socketclass = env->FindClass("java/net/Socket");
   jmethodID socketinit = env->GetMethodID(socketclass, "<init>", 
                                           "(Ljava/net/SocketImpl;)V");
   return env->NewObject(socketclass, socketinit, impl);
}

JNIEXPORT jobject JNICALL Java_FMG_IPcept_SocketControl_getLocalAddress
  (JNIEnv *env, jclass thisclass, jobject socket) {

   struct sockaddr_in sa;
   socklen_t len = sizeof(sa);

   int fd = getFDfromSocket(env, socket);

   if (getsockname(fd, (struct sockaddr *) &sa, &len) < 0) {
      /* this shouldn't happen */
      perror("getsockname(getLocalAddress)");
      _exit(1);
   }

   jclass iaclass = env->FindClass("java/net/InetAddress");
   jmethodID iainit = env->GetMethodID(iaclass, "<init>", "()V");
   jobject ia = env->NewObject(iaclass, iainit);

   jfieldID addrID = env->GetFieldID(iaclass, "address", "I");
   env->SetIntField(ia, addrID, ntohl(sa.sin_addr.s_addr));

   return ia;
}

JNIEXPORT jint JNICALL Java_FMG_IPcept_SocketControl_getLocalPort
  (JNIEnv *env, jclass thisclass, jobject socket) {

   struct sockaddr_in sa;
   socklen_t len = sizeof(sa);

   int fd = getFDfromSocket(env, socket);

   if (getsockname(fd, (struct sockaddr *) &sa, &len) < 0) {
      /* this shouldn't happen */
      perror("getsockname(getLocalPort)");
      _exit(1);
   }

   return ntohs(sa.sin_port);
}

JNIEXPORT jobject JNICALL Java_FMG_IPcept_SocketControl_getPeerAddress
  (JNIEnv *env, jclass thisclass, jobject socket) {

   struct sockaddr_in sa;
   socklen_t len = sizeof(sa);

   int fd = getFDfromSocket(env, socket);

   if (getpeername(fd, (struct sockaddr *) &sa, &len) <0) {
      jclass excl = env->FindClass("java/io/IOException");
      env->ThrowNew(excl, "getpeername() failed, unconnected socket?");
      return NULL;
   }

   jclass iaclass = env->FindClass("java/net/InetAddress");
   jmethodID iainit = env->GetMethodID(iaclass, "<init>", "()V");
   jobject ia = env->NewObject(iaclass, iainit);

   jfieldID addrID = env->GetFieldID(iaclass, "address", "I");
   env->SetIntField(ia, addrID, ntohl(sa.sin_addr.s_addr));

   return ia;
}

JNIEXPORT jint JNICALL Java_FMG_IPcept_SocketControl_getPeerPort
  (JNIEnv *env, jclass thisclass, jobject socket) {

   struct sockaddr_in sa;
   socklen_t len = sizeof(sa);

   int fd = getFDfromSocket(env, socket);

   if (getpeername(fd, (struct sockaddr *) &sa, &len) < 0) {
      jclass excl = env->FindClass("java/io/IOException");
      env->ThrowNew(excl, "getpeername() failed, unconnected socket?");
      return 0;
   }

   return ntohs(sa.sin_port);
}
